import requests
from requests.adapters import SSLError
import warnings
from pandas import concat
from bs4 import BeautifulSoup
from silverware import read_table
from ravenclaw import standardize_columns
from chronometry.progress import ProgressBar
from chronometry import get_elapsed, get_now
from time import sleep
from .countries import COUNTRIES as _COUNTRIES

NUM_RECORDS_PER_PAGE = 5000


class GuruFocus:
	def __init__(self, rate_limit_wait_seconds=0.01, cache=None, num_request_tries=4):
		self._rate_limit_wait = rate_limit_wait_seconds
		self._rate_limit_last_call = None
		self._num_request_tries = num_request_tries

		self._cache = cache
		if self._cache:
			self.request = self._cache.make_cached(
				id='gurufocus_request',
				function=self._request,
				sub_directory='gf_request'
			)
			self.get_table = self._cache.make_cached(
				id='gurufocus_get_table',
				function=self._get_table,
				sub_directory='gf_get_table'
			)

		else:
			self.request = self._request
			self.get_by_country = self._get_by_country
			self.get_table = self._get_table

	COUNTRIES = _COUNTRIES

	@property
	def cache(self):
		"""
		:rtype: disk.Cache or NoneType
		"""
		return self._cache

	def _request(self, url):
		for i in range(1, self._num_request_tries + 1):
			try:
				if self._rate_limit_wait and self._rate_limit_last_call:
					wait_time = self._rate_limit_wait - get_elapsed(start=self._rate_limit_last_call, unit='s')
					if wait_time > 0:
						sleep(wait_time)
				result = requests.get(url)
				break
			except SSLError as e:
				print(f'try {i}, error with get request with url="{url}"')
				warnings.warn(str(e))
				sleep(0.001 * 10 ** i)
		else:
			raise e

		if self._rate_limit_wait:
			self._rate_limit_last_call = get_now()

		return result

	@staticmethod
	def get_request_url(country, n, page=None):
		if page is None:
			return f'https://www.gurufocus.com/stock_list.php?n={n}&r={country}'
		else:
			return f'https://www.gurufocus.com/stock_list.php?n={n}&r={country}&p={page}'

	def _get_table(self, url):
		request = self.request(url)
		content = request.content
		soup = BeautifulSoup(content, "lxml")
		html_table = soup.find(name='table', attrs={'id': 'R1', 'class': 'R5'})
		table = read_table(html_table, parse_links=True, base_url='https://www.gurufocus.com')
		return standardize_columns(
			table.rename(columns={'': 'summary'})
		)

	def _get_by_country(self, country, n=None, page=None, echo=1):
		test = self.request(self.get_request_url(country=country, n=5)).content
		zzzz = self.request(self.get_request_url(country='ZZZZ', n=5)).content
		test_soup = BeautifulSoup(test, "lxml")
		zzzz_soup = BeautifulSoup(zzzz, "lxml")
		test_body = test_soup.find(name='table', attrs={'id': 'R1', 'class': 'R5'}).find(name='tbody')
		zzzz_body = zzzz_soup.find(name='table', attrs={'id': 'R1', 'class': 'R5'}).find(name='tbody')
		if test_body == zzzz_body:
			raise ValueError(country)

		test_page_links = test_soup.find(name='div', attrs={'class': 'page_links'})
		if test_page_links.text.lower().startswith('total records'):
			total_records = int(test_page_links.find(name='strong').text)
		else:
			raise ValueError('missing total records')

		if n is not None:
			url = self.get_request_url(country=country, n=n, page=page)
			return self.get_table(url=url)

		else:
			records = 0
			result = []
			this_page = None
			if total_records > NUM_RECORDS_PER_PAGE:
				progress_bar = ProgressBar(total=total_records, echo=echo)
			else:
				progress_bar = None

			while records < total_records:
				if progress_bar is not None:
					progress_bar.show(amount=records, text=f'{country}: {records} records')
				data = self._get_by_country(country=country, n=NUM_RECORDS_PER_PAGE, page=this_page)
				if len(data) < 1:
					break
				result.append(data)
				records += len(data)
				if page is None:
					page = 1
				else:
					page += 1
			if progress_bar is not None:
				progress_bar.show(amount=records, text=f'{country}: {records} records collected!')
			return concat(result).reset_index(drop=True)

	def get(self, country=None, n=None, echo=2):
		if country is not None:
			return self._get_by_country(country=country, n=n, echo=echo)
		else:
			progress_bar = ProgressBar(total=len(self.COUNTRIES), echo=echo)
			amount = 0
			results = []
			for country in self.COUNTRIES:
				progress_bar.show(amount=amount, text=country)
				result = self.get(country=country, n=n, echo=progress_bar - 1)
				result['country'] = country
				amount += 1
				results.append(result)
			results_data = concat(results).reset_index(drop=True)
			progress_bar.show(amount=amount, text=f'{len(results_data)} records collected!')
			return results_data



from selenium.webdriver.chrome.webdriver import WebDriver as ChromeDriver
from selenium.webdriver.firefox.webdriver import WebDriver as FirefoxDriver
from selenium.common.exceptions import TimeoutException
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions
from selenium.webdriver.common.by import By
from selenium.webdriver import Chrome, Firefox
#import urllib.request
import requests
from requests.adapters import SSLError
from bs4 import BeautifulSoup

from json import load as load_json

import time
import warnings
from chronometry import MeasurementSet


class Navigator:
	def __init__(self, driver=None, user_agent=None, request_method='urllib', timeout=10):
		"""
		:type driver: str or ChromeDriver or FirefoxDriver or NoneType
		:param str user_agent: the default user agent, one of random, ie, ff, chrome, etc.
		"""
		if isinstance(driver, str):
			if driver.lower() == 'chrome':
				self._driver = Chrome()
			elif driver.lower() == 'firefox':
				self._driver = Firefox()
			else:
				raise ValueError(f'Unknown driver: "{driver}"')
		else:
			if driver is None and request_method != 'urllib':
				raise ValueError(f'Driverless Navigator only works with urllib method, not with "{request_method}"!')
			self._driver = driver

		try:
			self._user_agent = user_agent.random
		except:
			self._user_agent = 'Mozilla/5.0 (Linux; Android 6.0; Nexus 5 Build/MRA58N) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/73.0.3683.103 Mobile Safari/537.36'

		self._default_request_method = request_method
		self._timeout = timeout
		self._url = None
		self._page_source = None
		self._parsed_html = None
		self._measurement_set = MeasurementSet()

		#self.get = self._measurement_set.measure(function=self._get, name='get')
		#self.parse_html = self._measurement_set.measure(function=self._parse_html, name='parse_html')
		self.get = self._get
		self.parse_html = self._parse_html

	def __del__(self):
		self.driver.quit()

	@property
	def driver(self):
		"""
		:rtype: ChromeDriver or FirefoxDriver
		"""
		return self._driver

	def _get_by_driver(self, url, element_id=None, timeout_exception='error'):
		self.driver.get(url=url)
		if element_id is None:
			time.sleep(self._timeout)
		else:
			try:
				element_present = expected_conditions.presence_of_all_elements_located((By.ID, element_id))
				WebDriverWait(driver=self.driver, timeout=self._timeout).until(element_present)

			except TimeoutException:
				self._page_source = None
				if timeout_exception[0].lower == 'e':
					raise TimeoutException(f'Timed out waiting for page:"{url}" to load!')
				elif timeout_exception[0].lower == 'w':
					warnings.warn(message=f'Timed out waiting for page:"{url}" to load!')
		return self.driver.page_source

	def _get_by_urllib(self, url, headers=None, json=False):
		headers = headers or {
			'user-agent': self._user_agent
		}
		request = requests.get(url=url, headers={'User-Agent': self._user_agent})
		'''
		request = urllib.request.Request(url)
		
		#for key, value in headers.items():
			request.add_header(key=key, val=value)

		if json:
			response = urllib.request.urlopen(request, timeout=self._timeout)
			return load_json(response)
		else:
			with urllib.request.urlopen(request, timeout=self._timeout) as response:
				html = response #.read().decode(encoding)
			return html
		'''
		if format == 'json':
			try:
				r = requests.get(url, params=parameters, headers=headers)
				result = r.json()
			except SSLError as e:
				warnings.warn(str(e))
				result = None

	def _get(
			self, url, request_method=None, element_id=None, get_json_back=False,
			timeout_exception='error', parser='lxml'
	):
		"""
		:type url: str
		:param str or NoneType request_method: one of 'urllib' or 'selenium' or None, None will choose the default
		:param str or NoneType user_agent: one of None (to choose default), random, ie, ff, etc.
		:type element_id: str or NoneType
		:type encoding: str
		:rtype: BeautifulSoup or str
		"""
		self._url = url
		request_method = request_method or self._default_request_method

		if request_method.lower() == 'urllib':
			html = self._get_by_urllib(url=url, json=get_json_back)

		elif request_method.lower() == 'selenium':
			html = self._get_by_driver(url=url, element_id=element_id, timeout_exception=timeout_exception)

		else:
			raise ValueError(f'Unknown method: "{method}"!')

		self._page_source = html

		if parser:
			return self.parse_html(parser=parser, html=html)

		else:
			return self._page_source

	def _parse_html(self, html, parser='lxml'):
		"""
		:type parser: str
		:type html: str
		:rtype: BeautifulSoup
		"""
		self._parsed_html = BeautifulSoup(html, parser)
		return self._parsed_html

	@property
	def loading_time(self):
		return self._get_timer.get_mean_elapsed(unit='sec')

	@property
	def parsing_time(self):
		return self._parse_timer.get_mean_elapsed(unit='sec')



COUNTRIES = [
	'USA', 'CAN', 'GBR', 'DEU', 'FRA', 'POL', 'RUS', 'SWE',
	'TUR', 'BIH', 'ITA', 'LUX', 'CHE', 'BEL', 'GRC', 'NOR',
	'ESP', 'DNK', 'BGR', 'NLD', 'ROU', 'FIN', 'SRB', 'AUT',
	'HRV', 'PRT', 'SVK', 'CYP', 'MKD', 'SVN', 'UKR', 'HUN',
	'LTU', 'LVA', 'MLT', 'ISL', 'EST', 'CZE', 'JPN', 'KOR',
	'THA', 'HKG', 'CHN', 'TWN', 'MYS', 'SGP', 'ISR', 'IDN',
	'VNM', 'PHL', 'LKA', 'BGD', 'KWT', 'NPL', 'SAU', 'JOR',
	'OMN', 'IRN', 'IRQ', 'ARE', 'KAZ', 'BAH', 'AUS', 'NZL',
	'BRA', 'MEX', 'ARG', 'CHL', 'PER', 'COL', 'PAN', 'JAM',
	'VEN', 'ECU', 'ZAF', 'EGY', 'NGA', 'MUS', 'MAR', 'ZWE',
	'KEN', 'TUN', 'GHA', 'CIV', 'IND', 'PAK'
]